//arenaManager.cpp
//Copyright © 1998, Paul Yu 

#include <stdio.h>
#include <opengl/GLView.h>
#include <MenuItem.h>
#include "arenaManager.h"
#include "message_protocols.h"

arenaManager::arenaManager() : BHandler("arenaManager")
{
	mView = NULL;

	mContinuous = false;
	mRepeatToken = 0;

	mSpinRate[0] = mSpinRate[1] = mSpinRate[2] = 0.;
	mDisplayAngle[0] = mDisplayAngle[1] =  mDisplayAngle[2] = 0;
}

arenaManager::~arenaManager()
{
	fprintf(stderr, "arenaManager::Destructor()\n");
	/*	*/
}

void arenaManager::MessageReceived(BMessage *msg)
{
	const uint32 REPEAT = 'rept';

	switch(msg->what)
	{
		case SET_CONTEXT:
		{
			BHandler *entity_manager;
			BGLView *view;
		
			if ((msg->FindPointer("client_view", (void **)&view) == B_OK) &&
				(msg->FindPointer("entity_manager", (void **)&entity_manager) == B_OK) &&
				entity_manager && view)
			{
				mView = view;
				mEntities = entity_manager;
				ResizeArena(25);

				BMessage token(REPEAT);
				token.AddInt32("token", ++mRepeatToken);
				Looper()->PostMessage(&token, this);	//post token back to itself
			}
		}
		break;

		case REPEAT:
		{
			if (msg->FindInt32("token") != mRepeatToken)
				return;
		
			if (mContinuous || Spinning())
			{
				if (mContinuous)
				{
					DisplayNext();
				}
				else
					Display();
			}
			else
				snooze(500000);

			//post token back to itself
			Looper()->PostMessage(msg, this);
		}
		break;

		case DRAW_ARENA:
		{
			if (mContinuous == false && Spinning() == false)
				Display();
		}
		break;

		case SINGLE_STEP:
		{
			if (mContinuous == false)
				DisplayNext();
		}
		break;

		case CONTINUOUS:
		{
	 		BMenuItem *item;
			mContinuous = !(mContinuous);
	 		msg->FindPointer("source", (void **)&item);
	 		item->SetMarked(mContinuous);
		}
		break;

		case RESTART:
		{
			BMessage reply;
			BMessenger messenger(mEntities);
			messenger.SendMessage(RESET, &reply);
			
			Display();
		}
		break;

	 	case X_SPIN_STOP:
	 	{
			mSpinRate[0] = 0.;
		}
		break;
	 
	 	case X_SPIN_POS:
	 	{	
			mSpinRate[0] += 1;	//Doesn't seem to spin at less than one degree
	 	}
 		break;

		case X_SPIN_NEG:
	 	{
			mSpinRate[0] -= 1;
	 	}
		break;
	 
	 	case Y_SPIN_STOP:
	 	{
			mSpinRate[1] = 0.;
		}
		break;
	 
	 	case Y_SPIN_POS:
	 	{	
			mSpinRate[1] += 1;
	 	}
 		break;

		case Y_SPIN_NEG:
	 	{
			mSpinRate[1] -= 1;
	 	}
		break;

	 	case Z_SPIN_STOP:
	 	{
			mSpinRate[2] = 0.;
		}
		break;
	 
	 	case Z_SPIN_POS:
	 	{	
			mSpinRate[2] += 1;
	 	}
 		break;

		case Z_SPIN_NEG:
	 	{
			mSpinRate[2] -= 1;
	 	}
		break;
				
		default:
			BHandler::MessageReceived(msg);
	}
}

void arenaManager::ResizeArena(int32 size)
{
	mArenaSize = size;

	mView->LockGL();

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glEnable(GL_DEPTH_TEST);
	glEnableClientState(GL_VERTEX_ARRAY);
	glShadeModel(GL_FLAT);
	glClearColor(0.0,0.0,0.0,0.0);
	glOrtho(-size,size,-size,size,-size,size);
	
	mView->UnlockGL();
}

void arenaManager::Display()
{
	if (Spinning())
		Rotate();

	{
		mView->LockGL();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glPushMatrix();
		
		// rotate for action!
		glRotatef(45, 1,1,1);
		glRotatef(mDisplayAngle[0], 1,0,0);
		glRotatef(mDisplayAngle[1], 0,1,0);
		glRotatef(mDisplayAngle[2], 0,0,1);
		
		// center the board
		float trans = mArenaSize/2.;
		glTranslatef(-trans,-trans,-trans);

		// ask it to draw the entities
		int32 gen_count;
		BMessage reply;
		BMessenger messenger(mEntities);
		messenger.SendMessage(DRAW_ENTITIES, &reply);
		gen_count = reply.FindInt32("generation");
		
		glPopMatrix();

		// status display	
		BView *v = mView->EmbeddedView();

		BFont font;
		font.SetSize(24.0);
		char status_display[32];
		sprintf(&status_display[0], "Generation: %d", (int)gen_count);
		
		//Lock the Window looper for the shortest possible time
		//to remain responsive to user input
		if(v != NULL && v->Window()->Lock())
		{
			v->SetFont(&font, B_FONT_SIZE);
			v->SetHighColor(0,255,0);
			v->SetDrawingMode(B_OP_OVER);
			v->DrawString(&status_display[0], BPoint(10,v->Bounds().bottom - 10));
			v->Sync();
			v->Window()->Unlock();
		}
		mView->SwapBuffers();
		mView->UnlockGL();
	}

	//don't skip frames
	if (mView->Looper()->Lock())
	{
		mView->Sync();
		mView->Looper()->Unlock();
	}
}

void arenaManager::DisplayNext()
{
	BMessage reply;
	BMessenger messenger(mEntities);

	//increment the cursor to the next generation
	messenger.SendMessage(INCREMENT, &reply);

	Display();
}

void arenaManager::Rotate()
{
	for(int32 i = 0; i < 3; i++)
	{
		if(mSpinRate[i] != 0.)
		{
			mDisplayAngle[i] += (int)mSpinRate[i];
			mDisplayAngle[i] = mDisplayAngle[i] % 360;
		}
	}
}

bool arenaManager::Spinning()
{
	return (mSpinRate[0] || mSpinRate[1] || mSpinRate[2]);
}
