//entityManager.cpp
//Copyright © 1998, Paul Yu 

#include <stdio.h>
#include "entityManager.h"
#include "message_protocols.h"

int32 BufferThread(void *data);

int32 BufferThread(void *data)
{
	((entityManager *)data)->Buffer();
	return 0;
}

entityManager::entityManager()
{
	mGeneration = 0;
	mBufferSize = 0;	//uninitialised
	mBufferCursor = 1;
	mDisplayCursor = 0;
	
	mReadyBuffers = create_sem(0, "consumer");
	mEmptyBuffers = create_sem(0, "producer");
	
	SetBufferSize(2);	//min. 2
	
	mBaseline = new BMessage((uint32)B_OK);
	
	Reset();

	mBufferThread = spawn_thread(&BufferThread, "Buffer Thread", B_LOW_PRIORITY, this);
	resume_thread(mBufferThread);
}

entityManager::~entityManager()
{
	status_t err;

	fprintf(stderr, "entityManager::Destructor()\n");
	
	//acquire all buffers
	int32 count = mBufferSize - 1;
	while(count)
	{	
		//bias towards acquiring empty buffers first
		if (acquire_sem_etc(mEmptyBuffers, 1, B_TIMEOUT, 10000) == B_OK)
			count--;
		else if (acquire_sem_etc(mReadyBuffers, 1, B_TIMEOUT, 10000) == B_OK)
			count--;
	}

	kill_thread(mBufferThread);
	wait_for_thread(mBufferThread, &err);

	for(int32 i = 0; i < mBufferSize; i++)
	{
		BList *list = (BList *)mBuffers.ItemAt(i);
		ClearBoard(list);
		delete list;
	}

	delete mBaseline;
}

void entityManager::MessageReceived(BMessage *msg)
{
	switch(msg->what)
	{
		case DRAW_ENTITIES:
		{
			Draw();
			
			if (msg->IsSourceWaiting())
			{
				BMessage reply((uint32)B_OK);
				reply.AddInt32("generation", mGeneration);
				msg->SendReply(&reply);
			}
		}
		break;
		
		case RESET:
		{
			Reset();
		}
		break;
		
		case INCREMENT:
		{
			Increment();
		}
		break;
		
		case SET_BUFFER_SIZE:
		{
			int32 size;
			if (msg->FindInt32("size", &size) == B_OK)
				SetBufferSize(size);
		}
		break;

		case SET_BASELINE:
		{
			delete mBaseline;
			mBaseline = new BMessage(msg);
			mBaseline->what = B_OK;
		}
		break;

		case GET_BASELINE:
		{
			if (msg->IsSourceWaiting())
			{
				msg->SendReply(mBaseline);
			}
		}
		break;

		case SET_CONTEXT:
		{
			addonManager *am;
			if ((msg->FindPointer("add-on_manager", (void **)&am) == B_OK) && am)
			{
				mAddons = am;
			}
		}
		break;
		
		default:
			BHandler::MessageReceived(msg);
	}
}

void entityManager::Reset()
{
	//acquire all buffers
	int32 count = mBufferSize - 1;
	while(count)
	{	
		//bias towards acquiring empty buffers first
		if (acquire_sem_etc(mEmptyBuffers, 1, B_TIMEOUT, 10000) == B_OK)
			count--;
		else if (acquire_sem_etc(mReadyBuffers, 1, B_TIMEOUT, 10000) == B_OK)
			count--;
	}

	mGeneration = 0;

	for(int32 i = 0; i < mBufferSize; i++)
	{
		BList *list = (BList *)mBuffers.ItemAt(i);
		ClearBoard(list);
	}

	mDisplayBoard = (BList *)mBuffers.ItemAt(0);
	BaselineToBoard(mDisplayBoard);

	mBufferCursor = 1;
	mDisplayCursor = 0;

	release_sem_etc(mEmptyBuffers, mBufferSize - 1, B_DO_NOT_RESCHEDULE);
}

void entityManager::Increment()
{
	acquire_sem(mReadyBuffers);

	mDisplayCursor = (mDisplayCursor + 1) % mBufferSize;
	mDisplayBoard = (BList *)mBuffers.ItemAt(mDisplayCursor);
	mGeneration++;

	release_sem_etc(mEmptyBuffers, 1, B_DO_NOT_RESCHEDULE);
}

//A circular buffer implementation
void entityManager::Buffer()
{
	BList all_children, entity_children;
	BList *prev, *next;
	BMessage *entity;
	const char *algorithm;
	int32 index, b;

	while(true)
	{
		acquire_sem(mEmptyBuffers);

		index = mBufferCursor - 1;
		if (index < 0)
			index = mBufferSize - 1;

		prev = (BList *)mBuffers.ItemAt(index);	
		next = (BList *)mBuffers.ItemAt(mBufferCursor);

		//Copy the previous generation to the next
		CopyBoard(prev, next);
			
		//Apply to behaviours to derive the BufferBoard
		index = 0;
		while((entity = (BMessage *)prev->ItemAt(index)) != NULL)
		{
			b = 0;
			entity_children.MakeEmpty();
			while((algorithm = entity->FindString("behaviours", b++)) != NULL)
				Behaviour(algorithm, index, prev, next, &entity_children);
			all_children.AddList(&entity_children);
			index++;
		}
	
		//Remove items that have died
		index = 0;
		while((entity = (BMessage *)next->ItemAt(index)) != NULL)
		{
			if (entity->FindBool("alive"))
				index++;
			else
			{
				next->RemoveItem((int32)index);
			}	
		}

		//Initialise and add Children
		index = 0;
		while((entity = (BMessage *)all_children.ItemAt(index++)) != NULL)
		{
			b = 0;
			while((algorithm = entity->FindString("initialisers", b++)) != NULL)
				Initialise(algorithm, entity);
		}
		next->AddList(&all_children);
		all_children.MakeEmpty();

		//move on to the next buffer
		mBufferCursor = (mBufferCursor + 1) % mBufferSize;

		release_sem(mReadyBuffers); //producer
	}
}

status_t entityManager::Draw()
{
	BMessage *entity;
	const char *algorithm;
	int32 count;
	int32 index = 0;
	int32 b;

	count = mDisplayBoard->CountItems();
	while((entity = (BMessage *)mDisplayBoard->ItemAt(index++)) != NULL)
	{
		b = 0;
		while((algorithm = entity->FindString("visuals", b++)) != NULL)
			Draw(algorithm, entity);
	}
	
	return B_OK;
}

//Dynamic resizing of the buffer pool -- hmm, useful?
void entityManager::SetBufferSize(int32 size)
{
	if (size < 2 || size == mBufferSize)
		return;

	//acquire all buffers
	int32 count = mBufferSize - 1;
	while(count > 0)
	{	
		//bias towards acquiring empty buffers first
		if (acquire_sem_etc(mEmptyBuffers, 1, B_TIMEOUT, 10000) == B_OK)
			count--;
		else if (acquire_sem_etc(mReadyBuffers, 1, B_TIMEOUT, 10000) == B_OK)
			count--;
	}
	
	//copy old pointers to temp, delete extra buffers
	BList temp;
	for(int32 i = 0; i < mBufferSize; i++)
	{
		int32 index = (mDisplayCursor + i) % mBufferSize;

		if (i < size)
		{
			temp.AddItem(mBuffers.ItemAt(index));
		}
		else
		{
			BList *list = (BList *)mBuffers.ItemAt(index);
			ClearBoard(list);
			delete list;
		}
	}

	mBuffers.MakeEmpty();
	mBuffers.AddList(&temp);

	//add new buffers if needed
	for(int32 i = mBufferSize; i < size; i++)
	{
		mBuffers.AddItem(new BList());
	}

	int32 ready, empty;
	if (mBufferCursor > mDisplayCursor)
		ready = (mBufferCursor - mDisplayCursor) - 1;
	else if	(mBufferCursor < mDisplayCursor)
		ready = mBufferCursor + (((mBufferSize - mDisplayCursor) - 1) - 1);
	else
		ready = mBufferSize - 1;
	
	if (ready >= size)
		ready = size - 1;
	
	empty = size - ready - 1;
		
	mDisplayCursor = 0;
	mBufferCursor = (ready + 1) % size;
	mBufferSize = size;

	//release Ready and Empty buffers
	release_sem_etc(mReadyBuffers, ready, B_DO_NOT_RESCHEDULE);
	release_sem_etc(mEmptyBuffers, empty, B_DO_NOT_RESCHEDULE);
}

void entityManager::BaselineToBoard(BList *board)
{
	BMessage entity;
	const char *algorithm;
	int32 index = 0, b;

	ClearBoard(board);

	while(mBaseline->FindMessage("entity", index++, &entity) == B_OK)
	{
		b = 0;
		while((algorithm = entity.FindString("initialisers", b++)) != NULL)
			Initialise(algorithm, &entity);
			
		board->AddItem(new BMessage(entity));
	}
}

status_t entityManager::Initialise(const char *algorithm, BMessage *entity)
{
	status_t (*function)(BMessage *);

	function = (status_t (*)(BMessage *))
		mAddons->GetInitialiser(algorithm);
	
	if (function)
		return function(entity);
	else
		return B_NAME_NOT_FOUND;
}

status_t entityManager::Behaviour(const char *algorithm, int32 index, BList *prev, BList *next, BList *children)
{
	status_t (*function)(const BList *, const BList *, int32, BList *);

	function = (status_t(*)(const BList *, const BList *, int32, BList *))
		mAddons->GetBehaviour(algorithm);
	
	if (function)
		return function(prev, next, index, children);
	else
		return B_NAME_NOT_FOUND;
}

status_t entityManager::Draw(const char *algorithm, BMessage *entity)
{
	status_t (*function)(BMessage *);

	function = (status_t (*)(BMessage *))
		mAddons->GetVisual(algorithm);
	
	if (function)
		return function(entity);
	else
		return B_NAME_NOT_FOUND;
}

void entityManager::ClearBoard(BList *board)
{
	int32 count = board->CountItems();
	for(int32 i = 0; i < count; i++)
	{
		delete (BMessage *)board->ItemAt(i);
	}
	
	board->MakeEmpty();
}


void entityManager::CopyBoard(BList *src, BList *dst)
{
	BMessage *entity;
	int32 index = 0;

	ClearBoard(dst);
	while((entity = (BMessage *)src->ItemAt(index++)) != NULL)
	{
		dst->AddItem(new BMessage(entity));
	}
}
